// Views.CheckXML.cs

// Version 1.28; 30 March 2006
// Copyright (C) 2004  Robert Nigel Horspool

/*

Views: Vendor Independent Extensible Windowing System
=====================================================

Queries to views@cs.up.ac.za
See http://views.cs.up.ac.za for
   - user and developer lists
   - new developments

A joint project of University of Victoria, Canada and
                    University of Pretoria, South Africa
Begun February 2002

Designed by
     Nigel Horspool (nigelh@uvic.ca)
     Judith Bishop (jbishop@cs.up.ac.za)

Original Windows implementation 2002-2004
   Nigel Horspool (nigelh@uvic.ca)

New implementation for Qt 2004
   Basil Worrall (bworrall@cs.up.ac.za)
   D-J Miller (anoyomouse@hotmail.com)

Port of Views on Qt to Linux
   University of Karlsruhe, June 2004
   Mihael Vrbanec (miqster@gmx.net)
   Johannes Eickhold (JEickhold@gmx.de)
   Matthias Kempka (mkempka@mkempka.de)

Initiated under the Microsoft Research Rotor RFP-1 Initiative

-----------

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

/* About version 1.24 of CheckXML
 * (This also applies to later versions, but this major change was made in 1.24)
 * 
 * This version of CheckXML re-factors the specification of supported controls
 * in such a way that the same version of CheckXML can be used by all platforms
 * that Views runs on.  Currently it is compatible with Views on WinForms and
 * Views on Qt, but can easily be extended to include other platforms should 
 * such be desired in the future.  The specification of controls is fairly 
 * simple, but flexible and powerful for this purpose, the details of it 
 * appear below.
 *
 * The variable platformCode must be set to the current platform, this is 
 * handled by adding compile time defines.  The default is 'W' for WinForms.  
 * To use 'Q' for Qt, the flag:
 *	/define:Qt
 * must be added to the command line for the compiler when the Views project 
 * is compiled.
 */

namespace Views 
{

using System;
using System.Text;
using System.Collections;


class CheckXML {
#if Qt
    // we will check assuming Qt library used for GUI rendering
    static char platformCode = 'Q';
#else
    // we assume Windows Forms library is used for GUI rendering
    static char platformCode = 'W';
#endif

    // The following table specifies which controls are supported in Views.
    // There is one entry per control, of the form
    //     tag: a1=t1 a2=t2 ...
    // where
    //     tag is the name of the control (in all lowercase letters),
    //     a1, a2 ... are the names of the attributes supported for the control
    //     t1, t2 ... are one letter codes representing the attribute type.
    //
    // By default, the tags and attributes are assumed to be supported in the
    // Views versions for all platforms.  However, prefixing a tag or prefixing
    // an attribute name with the annotation
    //     [W] or [W?]
    // means that the feature is supported when Windows Forms is used for GUI
    // rendering, and annotating with
    //     [Q] or [Q?]
    // means that the feature is supported when Qt us used for GUI rendering.
    // The difference between the notations, is that [W] and [Q] means that
    // execution is stopped when the feature is used on a platform which does
    // not support it, the addition of a ? as in [W?] or [Q?] means that the
    // feature is ignored on platforms that don't support it and execution
    // continues.
    //
    // Example 1:
    //     [W]groupbox: name=T font=F height=S
    // means that the GroupBox control would be supported only for the Windows
    // Forms platform (and would have just the three attributes listed).
    // Example 2:
    //     progressbar: name=T font=F [Q]height=S [Q]width=S
    // would mean that ProgressBar is supported in all versions of Views but
    // the height and width attributes would be supported only with Qt.
    // Example 3:
    //    vertical: [W?]halign=H [W?]valign=V
    // would mean that the halign and valign attributes are supported on
    // Windows and will be ignored if used on a Qt platform.
    //
    // The type codes for the attributes are as follows:
    //     A content alignment         K background colour (different default from C)
    //     B boolean                   N number (integer)
    //     C colour                    S size in points (integer)
    //     F font                      T text
    //     H horizontal alignment      V vertical alignment
    //     I image

    static string[] controlsSpec = new string[] {
	@"form: text=T debug=T font=F width=S height=S forecolor=C backcolor=K",

	@"button: name=T font=F height=S width=S text=T [W?]textalign=A
		forecolor=C backcolor=K [W?]halign=H [W?]valign=V image=I",

	@"openfiledialog: name=T font=F height=S width=S text=T [W?]textalign=A
		forecolor=C backcolor=K [W?]halign=H [W?]valign=V",

	@"savefiledialog: name=T font=F height=S width=S text=T [W?]textalign=A
		forecolor=C backcolor=K [W?]halign=H [W?]valign=V",

	@"checkbox: name=T font=F height=S width=S text=T [W?]textalign=A
		forecolor=C backcolor=K [W?]halign=H [W?]valign=V checked=B",

	@"radiobutton: name=T font=F height=S width=S text=T [W?]textalign=A
		forecolor=C backcolor=K [W?]halign=H [W?]valign=V checked=B",

	@"checkedlistbox: name=T font=F height=S width=S text=T [W?]textalign=A
		forecolor=C backcolor=K [W?]halign=H [W?]valign=V",

	@"combobox: name=T font=F height=S width=S text=T [W?]textalign=A
		forecolor=C backcolor=K [W?]halign=H [W?]valign=V",

	@"domainupdown: name=T font=F height=S width=S text=T [W?]textalign=H
		forecolor=C backcolor=K [W?]halign=H [W?]valign=V",

	@"textbox: name=T font=F height=S width=S text=T [W?]textalign=H
		forecolor=C backcolor=K [W?]halign=H [W?]valign=V readonly=B",

	@"listbox: name=T font=F height=S width=S text=T [W?]textalign=A
		forecolor=C backcolor=K [W?]halign=H [W?]valign=V",

	@"label: name=T font=F height=S width=S text=T [W?]textalign=A
		forecolor=C backcolor=K [W?]halign=H [W?]valign=V [Q]image=I",

	@"progressbar: name=T font=F height=S width=S text=T [W?]textalign=A
		forecolor=C backcolor=K [W?]halign=H [W?]valign=V
		minimum=N maximum=N value=N",

	@"trackbar: name=T font=F height=S width=S text=T [W?]textalign=A
		forecolor=C backcolor=K [W?]halign=H [W?]valign=V
		minimum=N maximum=N value=N tickfrequency=N largechange=N",

	@"picturebox: name=T [W?]font=F height=S width=S [W?]text=T [W?]textalign=A
		forecolor=C backcolor=K [W?]halign=H [W?]valign=V image=I",

	@"space: height=S width=S
		[W?]forecolor=C backcolor=K [W?]halign=H [W?]valign=V",

	@"vertical: height=S width=S
		[W?]forecolor=C backcolor=K [W?]halign=H [W?]valign=V",

	@"horizontal: font=F height=S width=S
		[W?]forecolor=C backcolor=K [W?]halign=H [W?]valign=V",

	@"[W]panel: name=T font=F height=S width=S forecolor=C
		backcolor=K halign=H valign=V",

	@"[W]groupbox: name=T font=F height=S width=S forecolor=C
		backcolor=K halign=H valign=V",

	@"[W]position: x=S y=S" };

    // All the error messages -- these are used as arguments to
    // ReportError which throws a Views exception
    static string badAlignmentMsg = "illegal alignment attribute {0}";
    static string badAttributeValueMsg = "{0} is an invalid value for {1} attribute";
    static string badDecimalMsg = "Numeric value expected, {0} found";
    static string badSpecStartMsg = "specification does not begin with <form> tag";
    static string badUnitsMsg = "Unknown measurement units ({0})";
    static string badXMLStartMsg = "Malformed XML opening to comment \"<!--\"";
    static string internalCodeErrorMsg = "Internal error -- unknown type code {0}";
    static string mismatchedTagsMsg = "close {0} not matched with open {1}";
    static string missingCloseQuoteMsg = "Missing close quote for attribute value";
    static string missingTagEndMsg = "Missing \">\" at end of tag";
    static string mustBeSingletonMsg = "{0} must be used as a singleton tag";
    static string negativeSizeMsg = "Negative sizes not allowed";
    static string nonexistentArgMsg = "Argument {0} is nonexistent";
    static string noPositionsMsg = "Empty <position>...</position> group";
    static string noRadioButtonsMsg = "Empty <RadioButton> list";
    static string notDecimalArgMsg = "{ddd} argument number is not a decimal integer";
    static string tooManyRadioButtonsMsg = "No more than one radio button in group can be on";
    static string unexpectedAttributeMsg = "Unexpected attribute {0}";
    static string unexpectedAttrsMsg = "Close tag cannot have attributes";
    static string unexpectedTagMsg = "unexpected control {0}";
    static string unrecognizedAttributeMsg = "unrecognized attribute {0} for {1}";
    static string unrecognizedTagMsg = "unrecognized tag {0}";
    static string unrecognizedXMLMsg = "unrecognized XML text";
    static string unsupportedAttributeMsg = "Attribute {0} of {1} not supported in this version of Views";
    static string unsupportedTagMsg = "tag {0} not supported in this version of Views";
    static string missingNameMsg = "Missing name attribute for {0} control";
    static string missingXorYMsg = "A position tag must include both X and Y attributes";

    static StringBuilder result;	// the normalized XML string is built in this variable
    static string x;			// the input XML string to be normalized
    static int ix;			// the current position in x
    static int len;			// the length of string x
    static int lineNum;			// the number of the line inside string x where ix refers
    static int lineStart;		// index of start of current line in string x
    static string[] args;		// arguments to substitute for {0} {1} ... inside x
    static Hashtable controls;		// table of supported controls and attributes

    // constructor -- it initializes the controls table
    static CheckXML() {
	char[] colon = {':'};
	char[] separators = {' ', '\n', '\t'};

	controls = new Hashtable();
	foreach(string s in controlsSpec) {
	    // extract the tagname
	    string[] parts = s.Split(colon);
	    string tagname = checkIfSupported(parts[0]);

	    Hashtable ht = new Hashtable();

	    string[] attrList = parts[1].Split(separators);
	    foreach(string pair in attrList) {
		int n = pair.IndexOf('=');
		if (n < 0) continue;
		string attr = checkIfSupported(pair.Substring(0,n));	// attribute name
		char code = pair[n+1];		// attribute's type code
         	ht.Add(attr,code);
	    } // foreach
	    controls.Add(tagname,ht);
	} // foreach

/*#if Debug
        foreach(string s in controls.Keys) {
            Console.WriteLine(s);
            ht = (Hashtable)(controls[s]);
            foreach (string t in ht.Keys) {
                Console.WriteLine("\t{0}={1}",t,ht[t]);
            }
        }
#endif*/
    }

    // Checks if the tag name is prefixed by the [Q],[Q?],[W],[W?] modifiers;
    // if the tag name is supported on our platform, name is the result;
    // otherwise name with "*" or "?" appended is the result.
    // The "*" means that use of the tag is fatal, whereas "?" means
    // that use of the tag is quietly ignored.
    private static string checkIfSupported( string name ) {
	if (name[0] != '[')
	    return name;
	// name is preceded by platform code(s)
	int i = name.IndexOf(']');
	string result = name.Substring(i+1);
	int j = name.IndexOf(platformCode);
	if (j >= 0 && j <= i)
	    return result;  // it's fully supported
	// check if attribute can be omitted
	j = name.IndexOf('?');
	result += (j < 0 || j > i)? "*" : "?";
	return result;
    }

    // Looks up a tag in the controls table and returns a list of the
    // attribute names and types which are allowed for that tag.
    private static Hashtable getKnownAttributes( string tag ) {
	Hashtable result = (Hashtable)(controls[tag]);
	if (result == null) {
	    result = (Hashtable)(controls[tag+"*"]);
	    if (result != null)
		ReportError(unsupportedTagMsg, tag);
	    result = (Hashtable)(controls[tag+"?"]);
	    if (result == null)
		ReportError(unrecognizedTagMsg, tag);
	}
	return result;
    }

    private enum TagType { open, close, singleton, other };

    // The XInfo class records info about an XML tag or text we have seen
    private class XInfo {
	public TagType t;		// kind of tag
	public string tag;		// tagname or just text if t == other
	public Hashtable attrs;		// table of attributes and values

	public override string ToString() {
	    switch(t) {
	    case TagType.open:       return "tag <"+tag+">";
	    case TagType.close:      return "tag </"+tag+">";
	    case TagType.singleton:  return "tag <"+tag+"/>";
	    default:                 return "text ("+tag+")";
	    }
	}
    }

    // Given a Views specification string (plus arguments) in the
    // relaxed XML notation, this function checks the string for
    // validity and returns a normalized version of the string
    // which can be processed by classes in the XML library.
    public static string Normalize( string x, params string[] args ) {
	result = new StringBuilder();
	CheckXML.x = x+" > \" \' -->";	// append some closing symbols
	CheckXML.ix = 0;
	CheckXML.args = args;
	CheckXML.lineNum = 1;
	CheckXML.lineStart = 0;
	len = x.Length;

	MatchForm();	// normalize top level

	string r = result.ToString();
	result = null;
	return r;
    }

    // Advances ix past white space and comments in string x
    private static void SkipSpaces() {
	bool inComment = false;
	for(;;) {
	    if (ix >= len) break;
	    char c = x[ix];
	    if (c == '<' && x[ix+1] == '!') {
		inComment = true;
		if (x[ix+2] != '-' || x[ix+3] != '-') {
		    ReportError(badXMLStartMsg);
		    break;
		}
		ix += 4;
		continue;
	    }
	    if (inComment) {
		if (c == '-' && x[ix+1] == '-' && x[ix+2] == '>') {
		    inComment = false;
		    ix += 3;
		    continue;
		}
	    } else {
		if (!Char.IsWhiteSpace(c)) break;
	    }
	    if (c == '\r') {
		if (x[ix+1] == '\n') ix++;
		lineNum++;
		lineStart = ++ix;
	    } else if (c == '\n') {
		lineNum++;
		lineStart = ++ix;
	    } else
		ix++;
	}
    }

    // Advances ix to start of next tag in string x
    private static void SkipText() {
	for(;;) {
	    if (ix >= len) break;
	    char c = x[ix];
	    if (c == '<') break;
	    ix++;
	}
    }

    // Extracts a tag name or attribute name from string x
    private static string GetName() {
	SkipSpaces();
	if (!Char.IsLetter(x[ix])) return null;
	int start = ix;
	for(;;) {
	    if (ix >= len || !Char.IsLetterOrDigit(x[ix])) break;
	    ix++;
	}
	return x.Substring(start,ix-start).ToLower();
    }

    // Extracts a value string which can appear after the equals
    // in an attribute=value pair in the string x
    private static string GetValue() {
	SkipSpaces();
	char delim = x[ix];

	if (delim == '{') {
	    ix++;
	    SkipSpaces();
	    int numPos = ix;
	    while(Char.IsDigit(x[ix]))
		ix++;
	    int num = Int32.Parse(x.Substring(numPos,ix-numPos));
	    SkipSpaces();
	    if (x[ix] != '}')
		ReportError(notDecimalArgMsg);
	    ix++;
	    if (num < 0 || num >= args.Length)
		ReportError(nonexistentArgMsg, num.ToString());
	    return args[num];
	}

	int start;
	if (delim == '\"' || delim == '\'' || delim == '') {
	    start = ++ix;
	} else {
	    delim = ' ';
	    start = ix;
	}
	for(;;) {
	    if (x[ix] == delim) break;
	    if (delim == ' ' && !Char.IsLetterOrDigit(x[ix])) {
		if (x[ix] != '.' && x[ix] != '-')
		    break;
	    }
	    if (delim == '' && x[ix] == '`') break;
	    ix++;
	}
	if (ix >= len) {
	    ix = start;
	    ReportError(missingCloseQuoteMsg);
	}
	if (delim == ' ')
	    return x.Substring(start,ix-start);
	ix++;
	return x.Substring(start,ix-start-1);
    }

    // Reads attribute=value pairs from the string x and returns
    // these pairs in the form of a hash table.
    private static Hashtable GetAttrs() {
	SkipSpaces();
	if (!Char.IsLetter(x[ix])) return null;
	Hashtable result = new Hashtable();
	string name, value;
	for(;;) {
	    name = GetName();
	    if (name == null) break;
	    SkipSpaces();  // spaces are illegal here
	    if (x[ix] != '=') {
		value = "";
	    } else {
		ix++;
		value = GetValue();
	    }
	    result[name] = value;
	}
	return result;
    }

    // Scans for tag ... open <X>, close </X> , singleton <X/>, or other
    // in string x and returns the tag as the result
    private static XInfo GetTag() {
	XInfo result = new XInfo();
	SkipSpaces();
	if (x[ix] != '<') {
	    result.t = TagType.other;
	    int start = ix;
	    SkipText();
	    result.tag = x.Substring(start,ix-start).Trim();
	    result.attrs = null;
	    return result;
	}
	ix++;
	if (x[ix] == '?') {
	    SkipText();
	    ix++;
	}

	SkipSpaces();  // spaces are illegal here
	if (x[ix] == '/') {
	    // close tag
	    result.t = TagType.close;
	    ix++;
	    result.tag = GetName();
	    result.attrs = GetAttrs();
	    if (result.attrs != null)
		ReportError(unexpectedAttrsMsg);
	    SkipSpaces();
	    if (x[ix] != '>' || ix >= len)
		ReportError(missingTagEndMsg);
	    else
		ix++;
	} else {
	    // open tag or singleton coming
	    result.t = TagType.open;
	    result.tag = GetName();
	    result.attrs = GetAttrs();
	    SkipSpaces();
	    if (x[ix] == '/' && x[ix+1] == '>') {
		result.t = TagType.singleton;
		ix += 2;
	    } else if (x[ix] == '>')
		ix++;
	    else
		ReportError(missingTagEndMsg);
	}
	if (controls[result.tag] == null && result.tag != "item") {
	    if (controls[result.tag+"*"] != null)
		ReportError(unsupportedTagMsg, result.tag);
	    if (controls[result.tag+"?"] == null)
		ReportError(unrecognizedTagMsg, result.tag);
	}
	return result;
    }

    // Checks and normalizes a <form>...</form> grouping
    private static void MatchForm() {
	XInfo firstTag = GetTag();
	if (firstTag.t != TagType.open || firstTag.tag != "form")
	    ReportError(badSpecStartMsg);
	XInfo nextTag;
	result.Append("<form");
	CheckAttributes(firstTag.tag, firstTag);
	result.Append(">");
	nextTag = GetTag();
	if (nextTag.t == TagType.open && nextTag.tag == "position") {
	    MatchPositionList("form", nextTag);
	    result.Append("</form>");
	    return;
	}
	for( ; ; ) {
	    switch(nextTag.t) {
	    case TagType.close:
	        if (nextTag.tag != firstTag.tag)
		    ReportError(mismatchedTagsMsg, nextTag, firstTag);
		result.Append("</form>");
		return;
	    case TagType.open:
		MatchGroup(nextTag);
		break;
	    case TagType.singleton:
		CheckSingleton(nextTag);
		break;
	    default:
		ReportError(unrecognizedXMLMsg);
		break;
	    }
	    nextTag = GetTag();
	}
	// not reached
    }

    // Checks and normalizes a singleton tag
    private static void CheckSingleton( XInfo c ) {
	if (c.tag == "panel" || c.tag == "groupbox") {
	    MatchGroup(c);
	} else {
	    result.Append("<"+c.tag);
	    CheckAttributes(c);
	    result.Append("/>");
	}
	if (c.t != TagType.singleton && c.t != TagType.open)
	    ReportError(unexpectedTagMsg, c);
    }

    // Invoked with firstTag being the first <position> tag in a list of
    // positions; context is either "panel" or "form"
    private static void MatchPositionList( string context, XInfo firstTag ) {
	while(firstTag.t == TagType.open && firstTag.tag == "position") {
	    MatchPosition(firstTag);
	    firstTag = GetTag();
	}
	if (firstTag.t != TagType.close || firstTag.tag != context) {
	    if (firstTag.t == TagType.close)
		ReportError(mismatchedTagsMsg, firstTag.tag, context);
	    else
		ReportError(unexpectedTagMsg, firstTag.tag);
	}
    }

    // Checks and normalizes a series of <RadioButton/> tags
    private static void MatchRadioButtonList( ) {
	XInfo c;
	int insertPos = 0;

	int checkedCount = 0;
	for( ; ; ) {
	    c = GetTag();
	    if (c.tag != "radiobutton") break;
	    if (c.t == TagType.open || c.t == TagType.singleton) { 
		result.Append("<radiobutton");
		CheckAttributes(c);
		foreach( Object key in c.attrs.Keys ) {
		    if ( ((string)key).ToLower().Equals("checked") ) {
			// if the checked attribute is true
			if (NormalizeAttr((string)(c.attrs[key]), 'B') == "1")
			    checkedCount++;
		    }
		}
		if (insertPos == 0)
		    insertPos = result.Length;
		result.Append("/>");
	    } else
		ReportError(unexpectedTagMsg, c.tag);
	    if (checkedCount > 1)
		ReportError(tooManyRadioButtonsMsg);
	}
	if (c.tag != "groupbox" || c.t != TagType.close)
	    ReportError(mismatchedTagsMsg, c.tag, "GroupBox");
	if (insertPos == 0)
	    ReportError(noRadioButtonsMsg);
	if (checkedCount == 0)
	    result.Insert(insertPos, " checked=\"1\"");
    }

    // Checks and normalizes a <X> ... </X> grouping
    private static void MatchGroup( XInfo firstTag ) {
	XInfo nextTag;

        if (controls[firstTag.tag] != null) {
	    result.Append("<"+firstTag.tag);
	    CheckAttributes(firstTag);
	    result.Append(">"); 

	    switch(firstTag.tag) {
		case "vertical":
		case "horizontal":
		    break;
		case "panel":
		    MatchPositionList("panel", GetTag());
		    result.Append("</panel>");
	    	    return;
		case "groupbox":
		    MatchRadioButtonList();
		    result.Append("</groupbox>");
	    	    return;
		case "checkedlistbox":
		case "combobox":
		case "domainupdown":
		    MatchItemList(firstTag);
		    result.Append("</"+firstTag.tag+">");
	    	    return;
		default:
		    nextTag = GetTag();
		    if (nextTag.t != TagType.close || nextTag.tag != firstTag.tag)
			ReportError(mustBeSingletonMsg, firstTag.tag);
	    	    return;
		}
	} else
	    ReportError(unrecognizedTagMsg, firstTag.tag);

	for(;;) {
	    nextTag = GetTag();
	    if (nextTag.t == TagType.close) break;
	    switch(nextTag.t) {
	    case TagType.open:
		MatchGroup(nextTag);
		break;
	    case TagType.singleton:
		CheckSingleton(nextTag);
		break;
	    default:
		ReportError(unrecognizedXMLMsg);
		break;
	    }
	}
	if (nextTag.tag != firstTag.tag)
	    ReportError(mismatchedTagsMsg, nextTag.tag, firstTag.tag);
	result.Append("</"+firstTag.tag+">");
    }

    // Checks and normalizes a series of <Item>...</Item> groups
    private static void MatchItemList( XInfo firstTag ) {
	XInfo nextTag;
	for(;;) {
	    nextTag = GetTag();
	    if (nextTag.t == TagType.close) break;
	    if (nextTag.t == TagType.open && nextTag.tag == "item") {
		result.Append("<item>");
		nextTag = GetTag();
		while(nextTag.t == TagType.other) {
		    result.Append(nextTag.tag);
		    nextTag = GetTag();
		}
		result.Append("</item>");
		if (nextTag.t == TagType.close && nextTag.tag == "item")
		    continue;
	    }
	    ReportError(unexpectedTagMsg, nextTag);
	}
	if (nextTag.tag != firstTag.tag)
	    ReportError(mismatchedTagsMsg, firstTag, nextTag);
    }

    // Checks and normalizes a <Position/> tag
    private static void MatchPosition( XInfo firstTag ) {
	result.Append("<position");
	CheckAttributes(firstTag);
	result.Append(">");		// bug correction (stmt was missing): 27-03-2006 (NH)
	XInfo nextTag = GetTag();
	switch(nextTag.t) {
	case TagType.close:
	    if (nextTag.tag != firstTag.tag)
		ReportError(mismatchedTagsMsg, firstTag, nextTag);
	    else
		ReportError(noPositionsMsg);
	    break;
	case TagType.open:
	case TagType.singleton:
	    CheckSingleton(nextTag);
	    break;
	default:
	    ReportError(unrecognizedXMLMsg);
	    break;
	}
	nextTag = GetTag();
	if (nextTag.t != TagType.close || nextTag.tag != "position") {
	    if (nextTag.t == TagType.close)
		ReportError(mismatchedTagsMsg, TagType.close, nextTag.t);
	    else
		ReportError(unexpectedTagMsg, nextTag.tag);
	}
	result.Append("</position>");
    } 

    // Checks if a=v is a valid setting where a is an alignment attribute
    static bool CheckAlignAttributes( string a, string v ) {
	bool isOK = false;
	string lv = v.ToLower();
	switch(a) {
	case "halign":
	    switch(lv) {
	    case "left":  case "right":  case "centre":  case "center":
		isOK = true;
		break;
	    }
	    break;
	case "valign":
	    switch(lv) {
	    case "top":  case "bottom":  case "middle":
	 	isOK = true;
		break;
	    }
	    break;
	case "textalign":
	    switch(lv) {
	    case "topleft":	case "topcenter":	case "topright":
	    case "middleleft":	case "middlecenter":	case "middleright":
	    case "bottomleft":	case "bottomcenter":	case "bottomright":
		isOK = true;
		break;	
	    }
	    break;
	default:
	    ReportError(unexpectedAttributeMsg, a);
	    return false;
	}
	if (!isOK)
	    ReportError(badAttributeValueMsg, v, a);
	return isOK;
    }

    // Checks if the attributes supplied for group g are legal
    private static void CheckAttributes( XInfo g ) {
	    CheckAttributes(g.tag, g);
    }

    // Checks if the attributes supplied for group g are legal
    private static void CheckAttributes( string controlName, XInfo g ) {
	Hashtable knownAttributes = getKnownAttributes(controlName);
	if (g.attrs == null) return;

	// some tags don't need a name attribute, for the rest it is required
	// only position needs an X and Y
	bool nameSeen = ( (g.tag == "form") || (g.tag == "label") ||
			  (g.tag == "image") || (g.tag == "space") ||
			  (g.tag == "vertical") || (g.tag == "horizontal") ||
			  (g.tag == "panel") || (g.tag == "position") ),
	     XSeen = (g.tag != "position"),
	     YSeen = (g.tag != "position");

	foreach( Object key in g.attrs.Keys ) {
	    string k = (string)key;
	    string v = (string)(g.attrs[k]);
	    if (v == null) v = "";

	    Object c = knownAttributes[k];
	    bool suppress = false;
	    if (c == null) {
		c = knownAttributes[k+"*"];
		if (c != null)
		    ReportError(unsupportedAttributeMsg, k, controlName);
		c = knownAttributes[k+"?"];
		if (c == null)
		    ReportError(unrecognizedAttributeMsg, k, g);
		suppress = true;
	    }
	    if (c != null) {
		v = NormalizeAttr(v, (char)c);
		if (!suppress)
		    result.Append(String.Format(" {0}=\"{1}\"", k, v));
	    }

	    if (k.ToLower().Equals("name") && v != null && v.Length > 0) {
		nameSeen = true;
	    }

	    if (k.ToLower().Equals("x")) XSeen = true;
	    if (k.ToLower().Equals("y")) YSeen = true;
	}

	if (!nameSeen)
	    ReportError(missingNameMsg, g.tag);

	if (!XSeen || !YSeen)
	    ReportError(missingXorYMsg);
    }

    static string[] colourNames = { // all colours known to System.Drawing.Color
	"AliceBlue", "AntiqueWhite", "Aqua", "Aquamarine", "Azure", "Beige",
	"Bisque", "Black", "BlanchedAlmond", "Blue", "BlueViolet", "Brown",
	"BurlyWood", "CadetBlue", "Chartreuse", "Chocolate", "Coral",
	"CornflowerBlue", "Cornsilk", "Crimson", "Cyan", "DarkBlue",
	"DarkCyan", "DarkGoldenrod", "DarkGray", "DarkGreen", "DarkKhaki",
	"DarkMagenta", "DarkOliveGreen", "DarkOrange", "DarkOrchid",
	"DarkRed", "DarkSalmon", "DarkSeaGreen", "DarkSlateBlue",
	"DarkSlateGray", "DarkTurquoise", "DarkViolet", "DeepPink",
	"DeepSkyBlue", "DimGray", "DodgerBlue", "Firebrick",
	"FloralWhite", "ForestGreen", "Fuchsia", "Gainsboro",
	"GhostWhite", "Gold", "Goldenrod", "Gray", "Green", "GreenYellow",
	"Honeydew", "HotPink", "IndianRed", "Indigo", "Ivory", "Khaki",
	"Lavender", "LavenderBlush", "LawnGreen", "LemonChiffon",
	"LightBlue", "LightCoral", "LightCyan", "LightGoldenrodYellow",
	"LightGreen", "LightGray", "LightPink", "LightSalmon",
	"LightSeaGreen", "LightSkyBlue", "LightSlateGray", "LightSteelBlue",
	"LightYellow", "Lime", "LimeGreen", "Linen", "Magenta", "Maroon",
	"MediumAquamarine", "MediumBlue", "MediumOrchid", "MediumPurple",
	"MediumSeaGreen", "MediumSlateBlue", "MediumSpringGreen",
	"MediumTurquoise", "MediumVioletRed", "MidnightBlue", "MintCream",
	"MistyRose", "Moccasin", "NavajoWhite", "Navy", "OldLace", "Olive",
	"OliveDrab", "Orange", "OrangeRed", "Orchid", "PaleGoldenrod",
	"PaleGreen", "PaleTurquoise", "PaleVioletRed", "PapayaWhip",
	"PeachPuff", "Peru", "Pink", "Plum", "PowderBlue", "Purple",
	"Red", "RosyBrown", "RoyalBlue", "SaddleBrown", "Salmon",
	"SandyBrown", "SeaGreen", "SeaShell", "Sienna", "Silver",
	"SkyBlue", "SlateBlue", "SlateGray", "Snow", "SpringGreen",
	"SteelBlue", "Tan", "Teal", "Thistle", "Tomato", "Turquoise",
	"Violet", "Wheat", "White", "WhiteSmoke", "Yellow", "YellowGreen" };

    // Checks and normalizes the attribute v which should have the typecode
    private static string NormalizeAttr(string v, char typecode) {
	string rv;
	int ix;
	switch(typecode) {
	    case 'A': // content alignment
		if (!CheckAlignAttributes("textalign",v))
		    ReportError(badAlignmentMsg, v);
		rv = v;
		break;

	    case 'B': // boolean
		switch(v[0]) {
		case '0':
		    int k = int.Parse(v);
		    rv = k==0? "0" : "1";
		    break;
		case 'f': case 'F': case 'n': case 'N':
		    rv = "0";
		    break;
		default:
		    rv = "1";
		    break;
		}
		break;

	    case 'C': // colour
		ix = Array.BinarySearch(colourNames, v, new CaseInsensitiveComparer());
		if (ix < 0) rv = "Black";
		else rv = colourNames[ix];
		break;

	    case 'F': // font
		rv = fixFontName(v);
		break;

	    case 'H': // horizontal alignment
		if (!CheckAlignAttributes("halign",v))
		    ReportError(badAlignmentMsg, v);
		rv = v;
		break;

	    case 'I': // image
		rv = v;
		break;
	
	    case 'K': // background colour
		ix = Array.BinarySearch(colourNames, v, new CaseInsensitiveComparer());
		if (ix < 0) rv = "LightPink";
		else rv = colourNames[ix];
		break;

	    case 'N': // number
		try {
		    int.Parse(v);
		} catch(Exception) {
		    ReportError(badDecimalMsg, v);
		}
		rv = v;
		break;
		
	    case 'S': // size
		double result = 0.0;
		int slen = v.Length;

		int i;
		for( i = 0;  i < slen;  i++ ) {
		    char c = v[i];
		    if (Char.IsDigit(c))
			continue;
		    if (c != '.' && c != ',')
			break;
		}
		result = Double.Parse(v.Substring(0,i));
		if (i < slen) {
		    string units = v.Substring(i);
		    switch(units.ToLower()) {
		    case "pt":  case "point":
			break;
		    case "in":  case "inch":
			result *= 72.0;  break;
		    case "mm":
			result *= (72.0/25.4);  break;
		    case "cm":
			result *= (72.0/2.54);  break;
		    case "pc":  case "pica":
			result *= 12.0;  break;
		    case "dd":  case "didot":
			result *= (72.0 * 0.01483);  break;
		    case "cc":  case "cicero":
			result *= (72.0 * 12.0 * 0.01483);  break;
		    default:
			ReportError(badUnitsMsg, units);
			break;
		    }
		}
		if (result < 0)
		    ReportError(negativeSizeMsg);
		// rv = result.ToString();
		// round to nearest integer
		rv = ((int)(result+0.5)).ToString();
		break;

	    case 'T': // text
		rv = v;
		break;

	    case 'V': // vertical alignment
		if (!CheckAlignAttributes("valign",v))
		    ReportError(badAlignmentMsg, v);
		rv = v;
		break;

	    default: 
		rv = v;
		ReportError(internalCodeErrorMsg, typecode);
		break;
	}
	return rv;
    }

    static string[,] fontBits =
	{{"bf", "bold", null, null},
	 {"md", "medium", null, null},
	 {"it", "italic", "em", "emphasis"},
	 {"up", "upright", null, null},
	 {"sf", "sansserif", "sans", null},
	 {"rm", "roman", null, null},
	 {"tt", "teletype", "courier", "monospace"}};

    private static string fixFontName( string v ) {
	string shape = "";
	string weight = "";
	string family = "";
	v = v.ToLower();
	double size = 10;
	while(v.Length > 0) {
	    string match = null;
	    if (char.IsLetter(v[0])) {
                for( int i=0;  i<fontBits.GetLength(0);  i++ ) {
                    for( int j=0;  j<fontBits.GetLength(1);  j++ ) {
                        string m = fontBits[i,j];
                        if (m == null) break;
                        if (v.StartsWith(m)) {
                            match = fontBits[i,0];
                            v = v.Substring(m.Length);
                            break;
                        }
                    }
                    if (match != null) break;
                }
                switch(match) {
                case "bf": case "md":
                    weight = match;
                    break;
                case "it": case "up":
                    shape = match;
                    break;
                case "rm": case "sf": case "tt":
                    family = match;
                    break;
                default:
                    v = v.Substring(1);
                    break;
                }
		continue;
	    }
	    if (char.IsDigit(v[0])) {
		bool dotSeen = false;
		int i;
		for( i=1; i < v.Length; i++ ) {
		    char c = v[i];
		    if (c == '.') {
			if (dotSeen) break;
			dotSeen = true;
			continue;
		    }
		    if (!Char.IsDigit(c)) break;
		}
		size = double.Parse(v.Substring(0,i));
		v = v.Substring(i);
	    }
	}
	return family+weight+shape+size;
    }

    private static void ReportError( string msg, params object[] args ) {
	throw new ViewsException(
	    "Error in XML Specification for Views.Form\n    "+
	    String.Format(" (at line {0}, column {1}):  {2}\n",
		lineNum, ix-lineStart+1, String.Format(msg,args)) +
	    "    " + x.Substring(lineStart, ix-lineStart) + "\n");
    }
} // class CheckXML

} // namespace Views
