// Views.Form.cs

// Version 1.28;  30 March 2006
// Copyright (C) 2004  Robert Nigel Horspool


/*
Views: Vendor Independent Extensible Windowing System
=====================================================

Queries to views@cs.up.ac.za
See http://views.cs.up.ac.za for
   - user and developer lists
   - new developments

A joint project of University of Victoria, Canada and
                    University of Pretoria, South Africa
Begun February 2002

Designed by
     Nigel Horspool (nigelh@uvic.ca)
     Judith Bishop (jbishop@cs.up.ac.za)

Original Windows implementation 2002-2003
   Nigel Horspool (nigelh@uvic.ca)

New implementation for Qt 2004
   Basil Worrall (bworrall@cs.up.ac.za)
   D-J Miller (anoyomouse@hotmail.com)

Port of Views on Qt to Linux
   University of Karlsruhe, June 2004
   Mihael Vrbanec (miqster@gmx.net)
   Johannes Eickhold (JEickhold@gmx.de)
   Matthias Kempka (mkempka@mkempka.de)

Initiated under the Microsoft Research Rotor RFP-1 Initiative

-----------

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

namespace Views {

using System;
using System.Drawing;
using System.Collections;
using System.ComponentModel;
using System.Data;
using System.IO;
using System.Xml;
using System.Threading;


public class Form : System.Windows.Forms.Form {
	public const string version = "Views.Form 1.28";
	public delegate void ExitHandler();
	public event ExitHandler Exit;
	private Hashtable controls = new Hashtable(10);
	private int tabix = 0;
	private Hashtable counter = new Hashtable(10);
	private ArrayList clickedList = new ArrayList(10);
	private string fileName = null;
	private Thread thisFormThread, constructorThread;
	private int verticalGap = 5;
	private int horizontalGap = 5;
	private int formWidth  = 300;
	private int formHeight = 300;
	private int panelNest = 0;
	private Hashtable fontcache = new Hashtable(5);

	private bool debug = false;
	private bool ignoreEvent = false;

	// activeForms is a list of all active instances of this class.
	// We use the list to avoid performing Dispose() twice on the same
	// instance and to know when the last instance is being closed.
	private static ArrayList activeForms = ArrayList.Synchronized(new ArrayList());
	private XmlNode root;


	public Form( string xmlspec, params string[] args ) {
		xmlspec = TryGetFromFile(xmlspec);
		string fixedXMLSpec = CheckXML.Normalize(xmlspec,args);
		Console.WriteLine("Normalized specification:\n{0}\n",fixedXMLSpec);
		XmlDocument spec = new XmlDocument();
		spec.LoadXml(fixedXMLSpec);
		if (debug)
			Console.WriteLine("Normalized specification:\n{0}\n",fixedXMLSpec);
		root = spec.DocumentElement;

		thisFormThread = new Thread(new ThreadStart(FormThreadMethod));
		thisFormThread.Start();

		constructorThread = Thread.CurrentThread;
		constructorThread.Suspend();
	}

	private void FormThreadMethod() {
		this.SuspendLayout();
		VGroup xg = ConstructDisplay(root);
		SetSizes(xg);
		SetPositions(xg, new Point(0,0));
		this.AutoScaleBaseSize = new System.Drawing.Size(5, 13);
		this.ClientSize = new System.Drawing.Size(formWidth, formHeight);
		this.ResumeLayout(false);

		constructorThread.Resume();
		activeForms.Add(this);  // keep track of active forms
		// Run will not return until some form of Exit event is flagged.
		// Since we are using the Dispose method to handle this eventuality,
		// a call is made there to Application.Exit.
		System.Windows.Forms.Application.Run(this);
	}

	// check if fn really is a filename and, if so, read the file
	private string TryGetFromFile( string fn ) {
		if (fn.Length == 0 || Char.IsWhiteSpace(fn[0]) || fn[0] == '<')
			return fn;  // not a valid filename
		StreamReader inStream;
		try {
			inStream = new StreamReader(fn);
		} catch( Exception ) {
			return fn;
		}
		string s;
		try {
			s = inStream.ReadToEnd();
			inStream.Close();
		} catch( Exception e ) {
			throw new ViewsException("Cannot read XML file",e);
		}
		return s;
	}

	/// <summary>
	/// Clean up any resources being used.
	/// </summary>
	protected override void Dispose( bool disposing ) {
		lock(this){
			Monitor.Pulse(this); // wake up ClickFunction if necessary
		}
		if (!activeForms.Contains(this)) return;  // already disposed, return
		activeForms.Remove(this);
		base.Dispose(disposing);
		if (Exit != null) Exit();

		/*if (activeForms.Count > 0)
			System.Windows.Forms.Application.ExitThread();
		else*/
			System.Windows.Forms.Application.Exit();
	}

	public void CloseGUI() {
		if (!activeForms.Contains(this)) return;  // already disposed, return
		/*if (activeForms.Count > 0)
			System.Windows.Forms.Application.ExitThread();
		else*/
			System.Windows.Forms.Application.Exit();
	}

	private string GetAttr(Hashtable attrs, string name, string deflt) {
		object ob = attrs[name];
		if (ob == null) return deflt;
		return ob.ToString();
	}

	private int GetAttr(Hashtable attrs, string name, int deflt) {
		object ob = attrs[name];
		if (ob == null) return deflt;
		float f = float.Parse((string)ob);
		return (int)(f+0.5);
	}

	// invoked with node being root of XML specification
	// result is a group of controls or subgroups
	private VGroup ConstructDisplay(XmlNode node) {
		VGroup result = new VGroup("form");

		if (node == null || node.Name.ToLower() != "form")
			throw new ViewsException("Top level tag must be <form>");
		this.Text = "Views Form";
		Hashtable attrs = GetNodeAttrs(node);
		this.Text = GetAttr(attrs, "text", "Views Form");
		debug = GetAttr(attrs, "debug", null) != null;
		for(XmlNode z = node.FirstChild; z != null; z = z.NextSibling) {
			Object g = (z.Name.ToLower() == "position")?
				PlacedControl(this, z) : ConstructGroup(this, z);
			if (g != null) result.m.Add(g);
		}
		copyWidthHeight(attrs, result);
		result.c = this;
		SetAttrs(this, "form", attrs);
		return result;
	}

	// Invoked with node being <vertical>, <horizontal>, <space>, <panel>
	// or any tag for a control
	private VControl ConstructGroup( System.Windows.Forms.Control owner, XmlNode node) {
		if (node == null)
			return null;
		string s = node.Name.ToLower();
		switch(s) {
		case "button":
		case "openfiledialog":
		case "savefiledialog":
		case "checkbox":
		case "checkedlistbox":
		case "combobox":
		case "domainupdown":
		case "textbox":
		case "listbox":
		case "label":
		case "progressbar":
		case "trackbar":
		case "picturebox":
		case "space":
			return ConstructControl(owner, node);
		case "panel":
			return ConstructPanel(owner, node);
		case "groupbox":
			return ConstructGroupBox(owner, node);
		case "vertical":
		case "horizontal":
			break;
		default:
			throw new ViewsException("Unexpected tag: "+s);
		}
		VGroup result = new VGroup(s);
		Hashtable attrs = GetNodeAttrs(node);
		copyWidthHeight(attrs, result);
		for(XmlNode z = node.FirstChild; z != null; z = z.NextSibling) {
			VControl g = ConstructGroup(owner, z);
			if (g != null) result.m.Add(g);
		}
		return result;
	}

	private void copyWidthHeight( Hashtable attrs, VControl result ) {
		float f;
		foreach( string s in attrs.Keys ) {
			string val = (string)(attrs[s]);
			if (debug) Console.WriteLine("* copyWidthHeight: {0}=\"{1}\"", s, val);
			switch(s) {
			case "width":
				f = float.Parse(val);
				result.ds.Width = (int)(f+0.5);
				break;
				// result.ds.Width = Convert.ToInt32(val);  break;
			case "height":
				f = float.Parse(val);
				result.ds.Height = (int)(f+0.5);
				break;
				// result.ds.Height = Convert.ToInt32(val);  break;
			case "halign":
				result.halign = val.ToLower();  break;
			case "valign":
				result.valign = val.ToLower();  break;
			}
		}
	}

	// invoked with node being a <position> tag
	private VControl PlacedControl( System.Windows.Forms.Control owner, XmlNode node ) {
		VControl result;
		if (node.Name.ToLower() != "position")
			throw new ViewsException("<position> tag expected");
		Hashtable attrs = GetNodeAttrs(node);
		XmlNode z = node.FirstChild;
		switch(z.Name.ToLower()) {
		case "groupbox":
			result = ConstructGroupBox(owner, z);
			break;
		case "panel":
			result = ConstructPanel(owner, z);
			break;
		default:
			result = ConstructControl(owner, z);
			break;
		}
		result.positionSet = true;
		Point np = new Point();
		np.X = GetAttr(attrs, "x", 0);
		np.Y = GetAttr(attrs, "y", 0);
		result.c.Location = np;
		return result;
	}

	private Hashtable GetNodeAttrs( XmlNode node ) {
		Hashtable attrs = new Hashtable(10);
		if (node.NodeType == XmlNodeType.Element) {
			foreach(XmlNode attrnode in node.Attributes)
				attrs[attrnode.Name.ToLower()] = attrnode.Value;
		}
		return attrs;
	}

	private VGroup ConstructGroupBox( System.Windows.Forms.Control owner, XmlNode node ) {
		VGroup result = new VGroup("groupbox");
		System.Windows.Forms.GroupBox gb = new System.Windows.Forms.GroupBox();
		result.c = gb;
		gb.Parent = owner;
		Hashtable attrs = GetNodeAttrs(node);
		copyWidthHeight(attrs, result);
		for(XmlNode z = node.FirstChild; z != null; z = z.NextSibling) {
			VControl g = ConstructControl(gb, z);
			if (g == null || g.tag != "radiobutton")
				throw new ViewsException("<RadioButton> expected");
			result.m.Add(g);
		}
		if (attrs["text"] == null) attrs["text"] = attrs["name"];
		SetAttrs(gb, "groupbox", attrs);
		owner.Controls.Add(gb);
		return result;
	}

	private VGroup ConstructPanel( System.Windows.Forms.Control owner, XmlNode node ) {
		VGroup result = new VGroup("panel");
		System.Windows.Forms.Panel panel = new System.Windows.Forms.Panel();
		result.c = panel;
		panel.BorderStyle = System.Windows.Forms.BorderStyle.Fixed3D;
		panel.AutoScroll = true;
		Hashtable attrs = GetNodeAttrs(node);
		copyWidthHeight(attrs, result);
		for(XmlNode z = node.FirstChild; z != null; z = z.NextSibling) {
			VControl g = PlacedControl(panel, z);
			if (g != null) result.m.Add(g);
		}
		SetAttrs(panel, "panel", attrs);
		owner.Controls.Add(panel);
		return result;
	}

	// invoked with node having the tag for a basic control
	private VControl ConstructControl( System.Windows.Forms.Control owner, XmlNode node ) {
		System.Windows.Forms.Control c;
		string ckd, imageFile;
		VControl result = new VControl(node.Name.ToLower());
		Hashtable attrs = GetNodeAttrs(node);
		string s = node.Name.ToLower();
		Font font = CreateFont((string)(attrs["font"]), owner.Font);
		string tmp;
		int height;
		System.Windows.Forms.Button but;
		string align = (string)(attrs["textalign"]);

		switch(s) {
		case "button":
			but = new System.Windows.Forms.Button();
			imageFile = (string)attrs["image"];
			if (imageFile != null) {
				but.Image =
					ExamineImageFile(imageFile, attrs, 100, 100);
			} else {
				if (attrs["text"] == null) attrs["text"] = attrs["name"];
				SetWidthHeightFromText(attrs,font);
			}
			if (align != null)
				but.TextAlign = lookupAlignment(align);
			c = but;
			c.Click += new System.EventHandler(this.ClickFunction);
			break;
		case "openfiledialog":
		case "savefiledialog":
			but = new System.Windows.Forms.Button();
			if (align != null)
				but.TextAlign = lookupAlignment(align);
			but.Click += new System.EventHandler(this.ClickFileDialog);
			if (attrs["text"] == null) attrs["text"] = attrs["name"];
			SetWidthHeightFromText(attrs,font);
			c = but;
			break;
		case "checkbox":
			System.Windows.Forms.CheckBox cb = new System.Windows.Forms.CheckBox();
			ckd = (string)(attrs["checked"]);
			if (ckd != null) {
				ckd = ckd.ToLower();
				if (ckd != "0" && ckd != "n") cb.Checked = true;
			}
			cb.AutoCheck = false;
			if (align != null)
				cb.TextAlign = lookupAlignment(align);
			cb.Click += new System.EventHandler(this.ClickFunction);
			if (attrs["text"] == null) attrs["text"] = attrs["name"];
			SetWidthHeightFromText(attrs,font);
			c = cb;
			break;
		case "radiobutton":
			System.Windows.Forms.RadioButton rb = new System.Windows.Forms.RadioButton();
			ckd = (string)(attrs["checked"]);
			if (ckd != null) {
				ckd = ckd.ToLower();
				if (ckd != "0" && ckd != "n") rb.Checked = true;
			}
			rb.AutoCheck = true;
			rb.CheckedChanged += new System.EventHandler(this.ClickFunction);
			if (align != null)
				rb.TextAlign = lookupAlignment(align);
			c = rb;
			if (attrs["text"] == null) attrs["text"] = attrs["name"];
			SetWidthHeightFromText(attrs,font);
			break;
		case "checkedlistbox":
			System.Windows.Forms.CheckedListBox clb = new System.Windows.Forms.CheckedListBox();
			for(XmlNode z = node.FirstChild; z != null; z = z.NextSibling) {
				// handle checkbox items
				tmp = z.Name.ToLower();
				if (tmp == "item") {
					for(XmlNode clba = z.FirstChild; clba != null; clba = clba.NextSibling) {
						if (clba.NodeType == XmlNodeType.Text)
							tmp = clba.Value;
						else if (clba.NodeType == XmlNodeType.CDATA)
							tmp = clba.Value;
					}
				} else {
					throw new ViewsException("Unexpected tag: "+tmp);
				}
				clb.Items.Add(tmp, false);
			}
			clb.ItemCheck += new System.Windows.Forms.ItemCheckEventHandler(ItemClickFunction);
			clb.CheckOnClick = false;
			c = clb;
			break;
		case "combobox":
			System.Windows.Forms.ComboBox cmb = new System.Windows.Forms.ComboBox();
			for(XmlNode z = node.FirstChild; z != null; z = z.NextSibling) {
				tmp = z.Name.ToLower();
				if (tmp == "item") {
					for(XmlNode item = z.FirstChild; item != null; item = item.NextSibling) {
						if (item.NodeType == XmlNodeType.Text)
							tmp = item.Value;
						else if (item.NodeType == XmlNodeType.CDATA)
							tmp = item.Value;
					}
				} else {
					throw new ViewsException("Unexpected tag: "+tmp);
				}
				cmb.Items.Add(tmp);
			}
			cmb.SelectedIndexChanged += new System.EventHandler(ClickFunction);
			c = cmb;
			break;
		case "domainupdown":
			System.Windows.Forms.DomainUpDown dud = new System.Windows.Forms.DomainUpDown();
			for(XmlNode z = node.FirstChild; z != null; z = z.NextSibling) {
				// handle checkbox items
				tmp = z.Name.ToLower();
				if (tmp == "item") {
					for(XmlNode clba = z.FirstChild; clba != null; clba = clba.NextSibling) {
						if (clba.NodeType == XmlNodeType.Text)
							tmp = clba.Value;
						else if (clba.NodeType == XmlNodeType.CDATA)
							tmp = clba.Value;
					}
				} else {
					throw new ViewsException("Unexpected tag: "+tmp);
				}
				dud.Items.Add(tmp);
			}
			//dud.SelectedIndex = 0;
			if (align != null)
				dud.TextAlign = lookupHAlignment(align);
			dud.SelectedItemChanged += new System.EventHandler(ClickFunction);
			c = dud;
			break;
		case "textbox":
			System.Windows.Forms.TextBox txb =
				new System.Windows.Forms.TextBox();
			if (attrs["width"]  == null) attrs["width"]  = "150";
			if (attrs["readonly"] != null) {
				if ((string)(attrs["readonly"]) != "0")
					txb.ReadOnly = true;
			}
			if (align != null)
				txb.TextAlign = lookupHAlignment(align);
			c = txb;
			break;
		case "listbox":
			System.Windows.Forms.ListBox lb = new System.Windows.Forms.ListBox();
			lb.HorizontalScrollbar = true;
			if (attrs["height"] == null) attrs["height"] = "250";
			if (attrs["width"]  == null) attrs["width"]  = "150";
			c = lb;
			break;
		case "label":
			System.Windows.Forms.Label lab = new System.Windows.Forms.Label();
			if (attrs["text"] == null) attrs["text"] = attrs["name"];
			SetWidthHeightFromText(attrs,font,1.0);
			// lab.AutoSize = true;
			if (align != null) {
				lab.TextAlign = lookupAlignment(align);
				result.alignmentSet = true;
			}
			c = lab;
			break;
		case "space":
			c = new System.Windows.Forms.Label();
			attrs["text"] = "";
			if (attrs["width"] == null) attrs["width"] = "1";
			if (attrs["height"] == null) attrs["height"] = "1";
			break;
		case "progressbar":
			System.Windows.Forms.ProgressBar pb =
				new System.Windows.Forms.ProgressBar();
			height = GetAttr(attrs,"height",50);
			if (attrs["height"] == null) attrs["height"] = height.ToString();
			if (attrs["width"] == null) attrs["width"] = (5*height).ToString();
			pb.Minimum = GetAttr(attrs,"minimum",0);
			pb.Maximum = GetAttr(attrs,"maximum",100);
			pb.Value   = GetAttr(attrs,"value",0);
			c = pb;
			break;
		case "trackbar":
			System.Windows.Forms.TrackBar tb =
				new System.Windows.Forms.TrackBar();
			height = GetAttr(attrs,"height",50);
			if (attrs["height"] == null) attrs["height"] = height.ToString();
			if (attrs["width"] == null) attrs["width"] = (5*height).ToString();
			tb.Minimum = GetAttr(attrs,"minimum",0);
			tb.Maximum = GetAttr(attrs,"maximum",100);
			tb.Value   = GetAttr(attrs,"value",0);
			tb.ValueChanged += new System.EventHandler(this.ClickFunction);
			c = tb;
			break;
		case "picturebox":
			System.Windows.Forms.PictureBox p =
					new System.Windows.Forms.PictureBox();
			p.SizeMode = System.Windows.Forms.PictureBoxSizeMode.StretchImage;
			imageFile = (string)(attrs["image"]);
			p.Image = ExamineImageFile(imageFile, attrs, 100, 100);
			c = p;
			break;
		default:
			throw new ViewsException("Unexpected tag: "+s);
		}
		result.halign = GetAttr(attrs, "halign", "left");
		result.valign = GetAttr(attrs, "valign", "top");
		SetAttrs(c, s, attrs);
		c.Font = font;
		owner.Controls.Add(c);
		result.c = c;
		return result;
	}

	private Image ExamineImageFile( string imageFile, Hashtable attrs,
			int defltWidth, int defltHeight ) {
		Image im = null;
		if (imageFile != null) {
		    try {
			im = Image.FromFile(imageFile);
		    } catch( Exception e ) {
			throw new ViewsException("Unable to read "+imageFile, e);
		    }
		}
		if (attrs == null) return im;
		if (attrs["height"] == null || attrs["width"] == null) {
		    int w, h;
		    if (im != null) {
			w = im.Width;    h = im.Height;
		    } else {
			w = defltWidth;  h = defltHeight;
		    }
		    if (attrs["height"] == null) {
			if (attrs["width"] == null) {
			    attrs["width"] = w.ToString();
			} else {
			    h = (h*Convert.ToInt32(attrs["width"]))/w;
			}
			attrs["height"] = h.ToString();
		    } else {
			w = (w*Convert.ToInt32(attrs["height"]))/h;
			attrs["width"] = w.ToString();
		    }
		}
		return im;
	}

	private void SetWidthHeightFromText( Hashtable attrs, Font font ) {
		SetWidthHeightFromText(attrs, font, 1.25);
	}

	private void SetWidthHeightFromText( Hashtable attrs, Font font, double scale ) {
		font = CreateFont((string)(attrs["font"]), font);
		if (attrs["width"] == null)
			attrs["width"] = (getWidth(font, (string)(attrs["text"]), scale) + 4).ToString();
		if (attrs["height"] == null)
			attrs["height"] = (font.GetHeight()+8).ToString();
	}

	private void SetAttrs( System.Windows.Forms.Control c, string s, Hashtable attrs ) {
		c.Width  = GetAttr(attrs, "width", 72);
		c.Height = GetAttr(attrs, "height", 20);
		c.TabIndex = ++tabix;
		c.Text = GetAttr(attrs, "text", "");
		string color = (String)(attrs["forecolor"]);
		if (color != null) c.ForeColor = Color.FromName(color);
		color = (String)(attrs["backcolor"]);
		if (color != null) c.BackColor = Color.FromName(color);
		c.Font = CreateFont((string)(attrs["font"]), c.Font);

		if (c is System.Windows.Forms.TrackBar) {
			System.Windows.Forms.TrackBar xc = (System.Windows.Forms.TrackBar)c;
			string tickFrequency = (String)(attrs["tickfrequency"]);
			if (tickFrequency != null) xc.TickFrequency = int.Parse(tickFrequency);
			string largeChange = (String)(attrs["largechange"]);
			if (largeChange != null) xc.LargeChange = int.Parse(largeChange);
		}

		string a = (String)(attrs["name"]);
		if (a == null) {
			Object ob = counter[s];
			int k = (ob == null)? 1 : (Int32)(ob) + 1;
			a = s + k;
			counter[s] = k;
		}
		c.Name = a;
		controls[a] = new VControl(s, c);
	}

	// Sets the sizes of all the controls represented by g
	private void SetSizes( VControl g ) {
		Size result = new Size(g.ds.Width, g.ds.Height);
		int code = 0;
		VControl xgc;
		switch(g.tag) {
		case "vertical":
		case "groupbox":
			code = 1;  break;
		case "horizontal":
			code = 2;  break;
		case "form":
			code = 3;  break;
		case "panel":
			xgc = (VControl)g;
			result.Width  = xgc.c.Width;
			result.Height = xgc.c.Height;
			code = 3;  break;
		default:	// a simple control
			xgc = (VControl)g;
			result.Width  = xgc.c.Width;
			result.Height = xgc.c.Height;
			break;
		}
		if (code > 0) {
			VGroup xg = (VGroup)g;
			Point np = new Point(0,0);
			for( int i=0; i<xg.m.Count; i++ ) {
				VControl v = (VControl)(xg.m[i]);
				SetSizes(v);
				if ((code & 1) == 1) {
					if (v.fs.Width > result.Width) result.Width = v.fs.Width;
					np.Y += v.fs.Height;
				}
				if ((code & 2) == 2) {
					if (v.fs.Height > result.Height) result.Height = v.fs.Height;
					np.X += v.fs.Width;
				}
			}
			// propagate width/height of widest/tallest to all in list
			for( int i=0; i < xg.m.Count; i++ ) {
				VControl v = (VControl)(xg.m[i]);
				if ((code & 1) == 1) v.fs.Width = result.Width;
				if ((code & 2) == 2) v.fs.Height = result.Height;
			}
			if ((code & 1) == 1) result.Height = Math.Max(result.Height,np.Y);
			if ((code & 2) == 2) result.Width  = Math.Max(result.Width, np.X);
			if (code == 3) {
				result.Height += verticalGap;
				result.Width += horizontalGap;
			}
			if (g.tag == "groupbox") {
				int verticalMargin;
				VGroup vg = (VGroup)g;
				if (vg.m.Count < 3) verticalMargin = vg.m.Count*10;
				else if (vg.m.Count < 6) verticalMargin = vg.m.Count*8;
				else verticalMargin = vg.m.Count*6;

				g.c.Width = result.Width + 5;
				result.Width =  g.c.Width - horizontalGap;

				/* This used to add 30 to result.Height, but it
				 * turns out a constant addition is inadequate.
				 * When I had 10 radio buttons only 9 were
				 * visible.  So I made this scale that I tested
				 * for 2,3,5,10,15 and 20 radio buttons, and it
				 * seems to work reasonably well for those
				 * values.
				 */
				g.c.Height = result.Height + verticalMargin;
				result.Height = g.c.Height - verticalGap;
			}
		}
		g.fs.Height = result.Height + verticalGap;
		g.fs.Width  = result.Width + horizontalGap;
		if (debug) Console.WriteLine("Views.Form: {0}: W={1}, H={2}",
				g.tag, result.Width, result.Height);
	}

	// Sets positions of all the controls represented by g
	//   p is the top-left coordinate where to start placement
	private void SetPositions( VControl g, Point p ) {
		if (debug)
			Console.WriteLine("Views.Form: {0}: p={1}", g.tag, p);
		Point np = new Point(p.X+horizontalGap/2, p.Y+verticalGap/2);
		int code = 0;

		switch(g.tag) {
		case "form":
			code = 3;  break;
		case "vertical":
			code = 1;  break;
		case "horizontal":
			code = 2;  break;
		default:	// a simple control or space, panel, or groupbox
			if (!g.positionSet) {
			    if (g.fs.Width > g.c.Width) {
				if (g.halign[0] == 'r')
					np.X += (g.fs.Width - g.c.Width);
				else if (g.halign[0] == 'c')
					np.X += (g.fs.Width - g.c.Width)/2;
			    }
			    if (g.fs.Height > g.c.Height) {
				if (g.valign[0] == 'b')
					np.Y += (g.fs.Height - g.c.Height);
				else if (g.valign[0] == 'm')
					np.Y += (g.fs.Height - g.c.Height)/2;
			    }
			    g.c.Location = np;
			    g.positionSet = true;
			    if (g.tag == "label" && !g.alignmentSet) {
				System.Windows.Forms.Label lab = (System.Windows.Forms.Label)(g.c);
				switch(g.halign[0]) {
				case 'r':
				    lab.TextAlign = ContentAlignment.TopRight;  break;
				case 'c':
				    lab.TextAlign = ContentAlignment.TopCenter;  break;
				}
			    }
			}
			if (panelNest == 0) {
			    int sz;
			    sz = p.X + g.c.Width + horizontalGap;
			    if (sz > formWidth)
				formWidth = sz;
			    sz = p.Y + g.c.Height + verticalGap;
			    if (sz > formHeight)
				formHeight = sz;
			}
			if (debug)
				Console.WriteLine("Views.Form: Tag {0} placed at {1}",
					g.tag, g.c.Location);
			if (g.tag == "panel") {
				code = 3;
				break;
			}
			if (g.tag == "groupbox") {
				code = 1;
				break;
			}
			return;
		}

		VGroup xg = (VGroup)g;
		if (code == 3) {
			np.X = horizontalGap/2;  np.Y = verticalGap/2;
		} else if (code == 1 && g.tag == "groupbox") {
			/* With 'form' and panels, points for child components
			 * are relative to that parent.  Groupboxes are the
			 * same way, and there was no firm starting point
			 * for np.X, just np.Y
			 * The 12 was picked to match the one given for Y,
			 * it's fairly arbitrary
			 */
			np.X = 12;
			np.Y = 12;  // leave space for groupbox label
		}
		if (g.tag == "panel")
		     panelNest++;
		for( int i=0; i<xg.m.Count; i++ ) {
			VControl v = (VControl)(xg.m[i]);
			int tempGBY = 0;
			if (v.tag == "groupbox")
				tempGBY = np.Y;
			SetPositions(v,np);
			if (code == 1) {  // vertical list or groupbox
				if (v.tag == "groupbox")
					np.Y = tempGBY + v.c.Height + verticalGap;
				else
					np.Y += v.fs.Height + verticalGap;
				if (np.Y > formHeight)
					formHeight = np.Y;
				if (v.fs.Width > formWidth)
					formWidth = v.fs.Width;
			} else
			if (code == 2) {  // horizontal list
				np.X += v.fs.Width  + horizontalGap;
				if (np.X > formWidth)
					formWidth = np.X;
				if (v.fs.Height > formHeight)
					formHeight = v.fs.Height;
			} else
			if (code == 3 && panelNest == 0) {  // panel or form
				if (v.fs.Width > formWidth)
					formWidth = v.fs.Width;
				if (v.fs.Height > formHeight)
					formHeight = v.fs.Height;
			}
		}
		if (g.tag == "panel")
		     panelNest--;
	}

	private int getWidth( Font f, string s, double scale ) {
		if (s == null) s = " ";
		SizeF w = GetAutoScaleSize(f);
		double mult = Math.Ceiling((w.Width*s.Length*scale)/72.0);
		int result = Convert.ToInt32(mult)*72;
		if (debug)
			Console.WriteLine("getWidth: scaleSize={0}, mult={1}, result={2}",
				w, mult, result);
		return result;
	}

	private ContentAlignment lookupAlignment( string val ) {
	    ContentAlignment result = ContentAlignment.MiddleLeft;
	    switch(val.ToLower()) {
	    case "topleft":
		result = ContentAlignment.TopLeft;  break;
	    case "topcenter":
		result = ContentAlignment.TopCenter;  break;
	    case "topright":
		result = ContentAlignment.TopRight;  break;
	    case "middleleft":
		result = ContentAlignment.MiddleLeft;  break;
	    case "middlecenter":
		result = ContentAlignment.MiddleCenter;  break;
	    case "middleright":
		result = ContentAlignment.MiddleRight;  break;
	    case "bottomleft":
		result = ContentAlignment.BottomLeft;  break;
	    case "bottomcenter":
		result = ContentAlignment.BottomCenter;  break;
	    case "bottomright":
		result = ContentAlignment.BottomRight;  break;
	    default:
		throw new ViewsException("Bad TextAlign value: "+val);
	    }
	    return result;
	}

	private System.Windows.Forms.HorizontalAlignment lookupHAlignment( string val ) {
	    System.Windows.Forms.HorizontalAlignment result;
	    switch(val.ToLower()) {
	    case "left":
		result = System.Windows.Forms.HorizontalAlignment.Left;  break;
	    case "centre":
	    case "center":
		result = System.Windows.Forms.HorizontalAlignment.Center;  break;
	    case "right":
		result = System.Windows.Forms.HorizontalAlignment.Right;  break;
	    default:
		throw new ViewsException("Bad horizontal alignment value: "+val);
	    }
	    return result;
	}


	// we assume the fontspec string has been normalized to be
        // family/weight/shape/size in that order.
	// Two letter codes that are understood come from LaTeX:
	// Family: rm = Roman, sf = SansSerif, tt = TeleType (monospace)
	// Weight: bf = bold face, md = medium
	// Shape:  up = upright, it = italic 
	// -- nulls (missing components) imply the same as the
	//    current font for the form.
	private Font CreateFont( string fontspec, Font fnt ) {
	    if (fontspec == null || fontspec == "") return fnt;
	    Font font = (Font)(fontcache[fontspec]);
	    if (font != null) return font;
	    FontFamily ff;
	    FontStyle fs = fnt.Style;
	    int pos = 0;
	    switch(fontspec.Substring(pos,2)) {
	    case "sf":
		ff = FontFamily.GenericSansSerif;  pos += 2; break;
	    case "tt":
		ff = FontFamily.GenericMonospace;  pos += 2; break;
	    case "rm":
		ff = FontFamily.GenericSerif;  pos += 2; break;
	    default:
		ff = fnt.FontFamily;  break;
	    }
	    switch(fontspec.Substring(pos,2))  { // weight
	    case "bf":
		fs |= FontStyle.Bold;  pos += 2;  break;
	    case "md":
		fs ^= (~FontStyle.Bold);  pos += 2;  break;
	    default:
		break;
	    }
	    switch(fontspec.Substring(pos,2)) { // shape
	    case "it":
		fs |= FontStyle.Italic;  pos += 2;  break;
	    case "up":
		fs ^= (~FontStyle.Italic);  pos += 2;  break;
	    default:
		break;
	    }
	    float size = (pos < fontspec.Length)?
		float.Parse(fontspec.Substring(pos)) : fnt.SizeInPoints;
	    font = new Font(ff, size, fs);
	    fontcache[fontspec] = font;
	    return font; 
	}

	private void ClickFunction(object sender, System.EventArgs e) {
		if (ignoreEvent) return;
		lock(this){
			string senderName = ((System.Windows.Forms.Control)sender).Name;
			clickedList.Add(senderName);
			// invert checkbox value
			if (sender is System.Windows.Forms.CheckBox) {
				System.Windows.Forms.CheckBox cb = (System.Windows.Forms.CheckBox)sender;
				cb.Checked = !cb.Checked;
			}
			if (clickedList.Count == 1)
				Monitor.Pulse(this);
		}
	}

	private void ItemClickFunction(object sender, System.Windows.Forms.ItemCheckEventArgs e) {
		lock(this){
			string senderName = ((System.Windows.Forms.Control)sender).Name;
			clickedList.Add(senderName);
			// invert checkbox value
			if (sender is System.Windows.Forms.CheckedListBox) {
				System.Windows.Forms.CheckedListBox clb = (System.Windows.Forms.CheckedListBox)sender;
				if (e.CurrentValue == System.Windows.Forms.CheckState.Unchecked)
					e.NewValue = System.Windows.Forms.CheckState.Checked;
				else
					e.NewValue = System.Windows.Forms.CheckState.Unchecked;
			}
			if (clickedList.Count == 1)
				Monitor.Pulse(this);
		}
	}

	private void ClickFileDialog(object sender, System.EventArgs e) {
		lock(this){
			System.Windows.Forms.FileDialog fd;
			string senderName = ((System.Windows.Forms.Control)sender).Name;
			//clickedName = senderName;

			VControl vc = LookUpControl(senderName);
			if (vc.tag == "openfiledialog")
				fd = new System.Windows.Forms.OpenFileDialog();
			else if (vc.tag == "savefiledialog")
				fd = new System.Windows.Forms.SaveFileDialog();
			else
				throw new ViewsException("Unexpected event: "+e.ToString());
			//fd.InitialDirectory = "c:\\" ;
			fd.Filter = "txt files (*.txt)|*.txt|All files (*.*)|*.*" ;
			fd.FilterIndex = 2 ;
			fd.RestoreDirectory = false ;

			if (fd.ShowDialog() != System.Windows.Forms.DialogResult.OK) {
				return;
			}
			clickedList.Add(senderName);
			fileName = fd.FileName;
			if (clickedList.Count == 1)
				Monitor.Pulse(this); 
		}
	}

	public string GetControl() {
		string result = null;
		lock(this){
			if (clickedList.Count == 0)
				Monitor.Wait(this);  // wait until a button is clicked
			// This is a special case for the terminating condition
			// it should only ever happen at the call to Pulse from
			// Dispose()
			if (clickedList.Count > 0)
			{
				result = (string)(clickedList[0]);
				clickedList.RemoveAt(0);
			}
		}
		return result;
	}

	//public string GetButton() {	// for backwards compatibility
	//	return GetControl();
	//}

	private VControl LookUpControl( string name ) {
		Object xc = controls[name];
		if (xc == null)
			throw new ViewsException("No such control named "+name);
		return (VControl)xc;
	}

	public string GetText( string name ) {
		VControl xc = LookUpControl(name);
		System.Windows.Forms.Control c = xc.c;
		switch(xc.tag) {
		case "openfiledialog":
		case "savefiledialog":
			return fileName;
		case "checkedlistbox":
			System.Windows.Forms.CheckedListBox clb = (System.Windows.Forms.CheckedListBox)c;
			IEnumerator myEnumerator = clb.CheckedIndices.GetEnumerator();
			if (myEnumerator.MoveNext())
				return clb.Items[(int)(myEnumerator.Current)].ToString();
			return "";
		case "combobox":
			System.Windows.Forms.ComboBox cmb = (System.Windows.Forms.ComboBox)c;
			if (cmb.SelectedIndex >= 0)
				return cmb.SelectedItem.ToString();
			return "";
		case "domainupdown":
			System.Windows.Forms.DomainUpDown dud = (System.Windows.Forms.DomainUpDown)c;
			if (dud.SelectedIndex >= 0)
				return dud.SelectedItem.ToString();
			return "";
		case "groupbox":
			System.Windows.Forms.GroupBox gb = (System.Windows.Forms.GroupBox)(xc.c);
			for( int i=0; i<gb.Controls.Count; i++ ) {
				System.Windows.Forms.Control v =
					(System.Windows.Forms.Control)(gb.Controls[i]);
				if (((System.Windows.Forms.RadioButton)v).Checked)
					return v.Text;
			}
			return "";  // no button checked?
		default:
			break;
		}
		return c.Text;
	}

	public string GetText( string name, int itemNum ) {
		VControl xc = LookUpControl(name);
		switch(xc.tag) {
		case "listbox":
			System.Windows.Forms.ListBox lb = (System.Windows.Forms.ListBox)(xc.c);
			if (itemNum >= 0 && itemNum < lb.Items.Count)
				return (lb.Items[itemNum]).ToString();
			break;
		default:
			return GetText(name);
		}
		return "";
	}

	public int GetValue( string name ) {
		VControl xc = LookUpControl(name);
		System.Windows.Forms.Control c = xc.c;
		switch(xc.tag) {
		case "checkedlistbox":
			System.Windows.Forms.CheckedListBox clb = (System.Windows.Forms.CheckedListBox)c;
			for( int i=0;  i < clb.Items.Count;  i++ ) {
				if (clb.GetItemCheckState(i) == System.Windows.Forms.CheckState.Checked)
					return i;
			}
			break;
		case "checkbox":
			return ((System.Windows.Forms.CheckBox)c).Checked? 1 : 0;
		case "radiobutton":
			return ((System.Windows.Forms.RadioButton)c).Checked? 1 : 0;
		case "combobox":
			return ((System.Windows.Forms.ComboBox)c).SelectedIndex;
		case "domainupdown":
			return ((System.Windows.Forms.DomainUpDown)c).SelectedIndex;
		case "progressbar":
			return ((System.Windows.Forms.ProgressBar)c).Value;
		case "trackbar":
			return ((System.Windows.Forms.TrackBar)c).Value;
		default:
			break;
		}
		// Console.WriteLine("tag = {0}, name={1}", xc.tag, name);
		return -1;
	}

	public int GetValue( string name, int itemNum ) {
		VControl xc = LookUpControl(name);
		switch(xc.tag) {
		case "checkedlistbox":
			System.Windows.Forms.CheckedListBox clb = (System.Windows.Forms.CheckedListBox)(xc.c);
			return clb.GetItemCheckState(itemNum) == System.Windows.Forms.CheckState.Checked? 1 : 0;
		case "combobox":
			System.Windows.Forms.ComboBox cmb = (System.Windows.Forms.ComboBox)(xc.c);
			return cmb.SelectedIndex == itemNum? 1 : 0;			
		case "domainupdown":
			System.Windows.Forms.DomainUpDown dud = (System.Windows.Forms.DomainUpDown)(xc.c);
			return dud.SelectedIndex == itemNum? 1 : 0;			
		default:
			return GetValue(name);
		}
	}

	public int GetValue( string name, string itemText ) {
		VControl xc = LookUpControl(name);
		switch(xc.tag) {
		case "checkedlistbox":
			System.Windows.Forms.CheckedListBox clb = (System.Windows.Forms.CheckedListBox)(xc.c);
	      		for( int i = 0;  i < clb.Items.Count;  i++ ) {
	      			if (clb.Items[i].ToString() == itemText)
	      				return clb.GetItemCheckState(i) == System.Windows.Forms.CheckState.Checked? 1 : 0;
	      		}
			break;
		case "combobox":
			System.Windows.Forms.ComboBox cmb = (System.Windows.Forms.ComboBox)(xc.c);
			if (cmb.SelectedIndex < 0) return 0;
			return cmb.Items[cmb.SelectedIndex].ToString() == itemText? 1 : 0;			
		case "domainupdown":
			System.Windows.Forms.DomainUpDown dud = (System.Windows.Forms.DomainUpDown)(xc.c);
			if (dud.SelectedIndex < 0) return 0;
			return dud.Items[dud.SelectedIndex].ToString() == itemText? 1 : 0;			
		default:
			return GetValue(name);
		}
		return -1;
	}

	public void PutText( string name, string cval ) {
		VControl xc = LookUpControl(name);
		System.Windows.Forms.Control c = xc.c;
		switch(xc.tag) {
		case "listbox":
			if (cval == null)
				((System.Windows.Forms.ListBox)c).Items.Clear();
			else
				((System.Windows.Forms.ListBox)c).Items.Add(cval);
			break;
		case "checkedlistbox":
			((System.Windows.Forms.CheckedListBox)c).Items.Add(cval, false);
			System.Windows.Forms.CheckedListBox clb = (System.Windows.Forms.CheckedListBox)c;
			for( int i = 0;  i < clb.Items.Count;  i++ ) {
				if (clb.Items[i].ToString() == cval)
					clb.SetItemChecked(i, false);
			}
			break;
		case "combobox":
			System.Windows.Forms.ComboBox cmb = (System.Windows.Forms.ComboBox)c;
			for( int i = 0;  i < cmb.Items.Count;  i++ ) {
				if (cmb.Items[i].ToString() == cval) {
					ignoreEvent = true;
					cmb.SelectedIndex = i;
					ignoreEvent = false;
					break;
				}
			}
			break;
		case "domainupdown":
			System.Windows.Forms.DomainUpDown dud = (System.Windows.Forms.DomainUpDown)c;
			for( int i = 0;  i < dud.Items.Count;  i++ ) {
				if (dud.Items[i].ToString() == cval) {
					ignoreEvent = true;
					dud.SelectedIndex = i;
					ignoreEvent = false;
					break;
				}
			}
			break;
		case "textbox":
		case "button":
		case "checkbox":
		case "radiobutton":
			c.Text = cval;
			break;
		default:
			throw new ViewsException("Cannot write text to "+name);
		}
	}

	public void PutText( string name, int itemNum, string cval ) {
		VControl xc = LookUpControl(name);
		switch(xc.tag) {
		case "listbox":
			if (itemNum < 0) return;
			System.Windows.Forms.ListBox lb = (System.Windows.Forms.ListBox)(xc.c);
			int cnt = lb.Items.Count;
			while(cnt <= itemNum) {
				lb.Items.Add("");  cnt++;
			}
                        if (cval == null)
                           lb.Items.Remove(itemNum);
                        else
			   lb.Items[itemNum] = cval;
			break;
		default:
			PutText(name,cval);
			break;
		}
	}

	public void PutImage( string name, Image im ) {
		VControl xc = LookUpControl(name);
		System.Windows.Forms.Control c = xc.c;
		switch(xc.tag) {
		case "button":
			((System.Windows.Forms.Button)c).Image = im;
			break;
		case "picturebox":
			((System.Windows.Forms.PictureBox)c).Image = im;
			break;
		default:
			throw new ViewsException("Image attribute for "
				+name+" not supported");
		}
	}

	public void PutImage( string name, string fname ) {
		PutImage(name, ExamineImageFile(fname, null, 0, 0));
	}

	public void PutValue( string name, int v ) {
		VControl xc = LookUpControl(name);
		System.Windows.Forms.Control c = xc.c;
		switch(xc.tag) {
		case "progressbar":
			System.Windows.Forms.ProgressBar pb = (System.Windows.Forms.ProgressBar)c;
			if (v < pb.Minimum) v = pb.Minimum;
			if (v > pb.Maximum) v = pb.Maximum;
			pb.Value = v;
			break;
		case "trackbar":
			System.Windows.Forms.TrackBar tb = (System.Windows.Forms.TrackBar)c;
			if (v < tb.Minimum) v = tb.Minimum;
			if (v > tb.Maximum) v = tb.Maximum;
			ignoreEvent = true;
			tb.Value = v;
			ignoreEvent = false;
			break;
		case "checkedlistbox":
			System.Windows.Forms.CheckedListBox clb = (System.Windows.Forms.CheckedListBox)c;
			clb.SetItemChecked(v, false);
			break;
		case "combobox":
			ignoreEvent = true;
			((System.Windows.Forms.ComboBox)c).SelectedIndex = v;
			ignoreEvent = false;
			break;
		case "domainupdown":
			ignoreEvent = true;
			((System.Windows.Forms.DomainUpDown)c).SelectedIndex = v;
			ignoreEvent = false;
			break;
		case "checkbox":
			System.Windows.Forms.CheckBox cb = (System.Windows.Forms.CheckBox)c;
			cb.Checked = (v != 0);
			break;
		default:
			throw new ViewsException("Cannot assign integer to "+name);
		}
	}

	public VControl this[ string name ] {
	    get {
		return LookUpControl(name);
	    }
	}

} // class Form

} // namespace Views
